// Test app for MySQL 4.1.23 embedded server
//
// This app demonstrates a memory leak in the MySQL embedded server
//  when prepared statements are re-used.
//
// The leak first appeared in MySQL 4.1.21, and still exists in MySQL 4.1.23,
//  on both Mac and Windows.
//
// The leak didn't exist in MySQL 4.1.20, and doesn't exist in MySQL 5.1.22rc.
//
// To reproduce the leak, link this test app against (say) MySQL 4.1.23,
//  with 'REUSE_PREPARED_STATEMENT' set to 1.  Then execute this app,
//  and watch its memory usage grow infinitely (using Task Manager or similar).
//
// To make the leak go away, either link against (say) MySQL 4.1.20 or 5.1.22, 
//  or set 'REUSE_PREPARED_STATEMENT' to 0.  Then execute this app,
//  and watch its memory usage remain low and stable.


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <mysql.h>

#define true 1
#define false 0


#define REUSE_PREPARED_STATEMENT 1  // to make the leak disappear, set this to zero

#define DB_NAME_STR "test_db"
#define BIND_BUF_SIZE 1024L


static char *server_options[] = 
{ 
	"mysql_test", "--skip-bdb", "--skip-innodb", "--datadir=./var"
};
int num_elements = sizeof(server_options)/ sizeof(char *);

int main(void)
{
	MYSQL *mysql1;

        printf("server init\n");
	mysql_server_init(num_elements, server_options, NULL);
        printf("mysql init\n");
	mysql1 = mysql_init(NULL);
        printf("real connect\n");
	mysql_real_connect(mysql1, NULL, NULL, NULL, NULL, 0, NULL, 0);

	printf("Create a test database.\n");
	mysql_query(mysql1, "CREATE DATABASE IF NOT EXISTS " DB_NAME_STR);
	mysql_query(mysql1, "USE " DB_NAME_STR);

	// Create a test table with three rows.
	mysql_query(mysql1,	"DROP TABLE IF EXISTS table1");
	mysql_query(mysql1, 
		"CREATE TABLE `table1` ("
		" `id` int unsigned NOT NULL auto_increment, "
		" `term` TEXT NOT NULL, "
		" PRIMARY KEY (`id`) "
		" ) "
		" TYPE=MyISAM"
		" CHARSET=utf8");
	mysql_query(mysql1, 
		"INSERT INTO `table1` (term) VALUES "
		"(\"term1\"),"
		"(\"term2\"),"
		"(\"term3\")");

#if !REUSE_PREPARED_STATEMENT // i.e. infinite loop which creates/executes/releases a prepared statement
	while (true)
#endif
	{
		// Create a prepared query.
		const char* command = "SELECT term FROM table1"; 
		MYSQL_STMT* stmt = mysql_stmt_init(mysql1); 
		mysql_stmt_prepare(stmt, command, (int) strlen(command)); 

	
#if REUSE_PREPARED_STATEMENT
		while (true) // i.e. infinite loop which always re-executes an existing prepared statement
#endif
		{
			// Initialize buffer to receive results of prepared query
			MYSQL_BIND	bind;
			char bindBuf[BIND_BUF_SIZE];
			unsigned long bindLen;
			my_bool bindIsNull;
			bind.buffer_type = MYSQL_TYPE_STRING;
			bind.buffer = &bindBuf;
			bind.buffer_length = BIND_BUF_SIZE;
			bindLen = bind.buffer_length;
			bind.length = &bindLen;
			bindIsNull = 0;
			bind.is_null = &bindIsNull;
			bind.is_unsigned = 1;

			// show current memory consumption
			// vsize is the 23rd field in Linux /proc/self/stat
			// if you know a better way to get this please let me know
			// note that getrusage() does *not* report this on Liunux
			// the rusage.ru_*rss fields in the rusage structure are
			// not filled by getrusage() on Linux
			{
				FILE *fp = fopen("/proc/self/stat", "r");
				if (fp) {
					char buf[1024], *p;
					int i;

					fgets(buf, 1024, fp);
					
				        p = strtok(buf, " ");

					for (i=1; i<23; i++) {
						p = strtok(NULL, " ");
						if (!p) break;
					}	
					if (p) printf("\r%s", p);
					
					fclose(fp);
				}
			}

			// Execute the prepared query
			mysql_stmt_execute(stmt); 
			mysql_stmt_bind_result(stmt, &bind);
			
			// Iterate through the query results
			int fetchCount = 0;
			while (!mysql_stmt_fetch(stmt)) 
				fetchCount++;

			// Release the query results
			mysql_stmt_free_result(stmt); 
		}

		// Release the query
		mysql_stmt_close(stmt);
	}

	mysql_close(mysql1);
	mysql_server_end();
	return 0;
}	

